try:
    from importlib.metadata import PackageNotFoundError, version
except ImportError:
    # for python < 3.8
    from importlib_metadata import PackageNotFoundError, version

try:
    __version__ = version(__package__)
except PackageNotFoundError:
    __version__ = "0+unknown"
